#include <stdlib.h>
#include <time.h>
#include <stdio.h>
#include <immintrin.h>

void array_sum_sequential ( int *a, int *b, int *c, int length ) {
    for ( int i = 0; i < length; ++i ) {
        c[i] = a[i] + b[i];
    }
}

void array_sum_vectorized ( int *a, int *b, int *c, int length ) {
    // __m256i can pack 8 ints
    int overflow = length % 8;
    
    for ( int i = 0; i < ( length - overflow ); i += 8 ) {
        __m256i a_vector = _mm256_loadu_si256 ( a + i );
        __m256i b_vector = _mm256_loadu_si256 ( b + i );
        
        __m256i c_vector = _mm256_add_epi32 ( a_vector, b_vector );
        
        _mm256_storeu_si256 ( c + i, c_vector );
    }
    
    for ( int i = length - overflow; i < length; ++i ) {
        c[i] = a[i] + b[i];
    }
}

int main ( int argc, char **argv ) {
    int type       = 1;
    int iterations = 10;
    int length     = 16;
    int min        = 0;
    int max        = 10;
    
    if ( argc >= 2 ) { type       = atoi ( argv[1] ); }
    if ( argc >= 3 ) { iterations = atoi ( argv[2] ); }
    if ( argc >= 4 ) { length     = atoi ( argv[3] ); }
    if ( argc >= 5 ) { min        = atoi ( argv[4] ); }
    if ( argc >= 6 ) { max        = atoi ( argv[5] ); }
    
    int *a = calloc ( length, sizeof ( int ) );
    int *b = calloc ( length, sizeof ( int ) );
    int *c = calloc ( length, sizeof ( int ) );
    
    if ( a == NULL || b == NULL || c == NULL ) {
        return 1;
    }
    
    for ( int i = 0; i < length; ++i ) {
        a[i] = ( int ) ( 1. * rand ( ) / RAND_MAX * ( max - min ) + min );
        b[i] = ( int ) ( 1. * rand ( ) / RAND_MAX * ( max - min ) + min );
    }
    
    clock_t start = clock ( );
    for ( int i = 0; i < iterations; ++i ) {
        switch ( type ) {
            case 0: {
                array_sum_sequential ( a, b, c, length );
                break;
            }
            case 1: {
                array_sum_vectorized ( a, b, c, length );
                break;
            }
        }
    }
    clock_t end = clock ( );
    
    double elapsed_in_ms = ( end - start ) * 1. / CLOCKS_PER_SEC * 1000;
    
    printf ( "ELAPSED (ms) = %lf\n", elapsed_in_ms );
    
    free ( a );
    free ( b );
    free ( c );
    
    return 0;
}