#include <stdlib.h>
#include <time.h>
#include <stdio.h>
#include <limits.h>
#include <immintrin.h>

int array_min_sequential ( int *a, int length ) {
    int min = INT_MAX;
    
    for ( int i = 0; i < length; ++i ) {
        if ( min > a[i] ) {
            min = a[i];
        }
    }
    
    return min;
}

int array_min_vectorized ( int *a, int length ) {
    // __m256i can pack 8 ints
    int overflow = length % 8;
    
    __m256i min_vector = _mm256_set1_epi32 ( INT_MAX );
    
    for ( int i = 0; i < ( length - overflow ); i += 8 ) {
        __m256i a_vector = _mm256_loadu_si256 ( a + i );
        
        min_vector = _mm256_min_epi32 ( a_vector, min_vector );
    }
    
    int tmp[8] = { 0 };
    
    _mm256_storeu_si256 ( tmp, min_vector );
    
    int min = INT_MAX;
    for ( int i = 0; i < 8; ++i ) {
        if ( min > tmp[i] ) {
            min = tmp[i];
        }
    }
    
    for ( int i = length - overflow; i < length; ++i ) {
        if ( min > a[i] ) {
            min = a[i];
        }
    }
    
    return min;
}

int main ( int argc, char **argv ) {
    int type       = 1;
    int iterations = 10;
    int length     = 16;
    int min        = 0;
    int max        = 10;
    
    if ( argc >= 2 ) { type       = atoi ( argv[1] ); }
    if ( argc >= 3 ) { iterations = atoi ( argv[2] ); }
    if ( argc >= 4 ) { length     = atoi ( argv[3] ); }
    if ( argc >= 5 ) { min        = atoi ( argv[4] ); }
    if ( argc >= 6 ) { max        = atoi ( argv[5] ); }
    
    int *a = calloc ( length, sizeof ( int ) );
    
    if ( a == NULL ) {
        return 1;
    }
    
    for ( int i = 0; i < length; ++i ) {
        a[i] = ( int ) ( 1. * rand ( ) / RAND_MAX * ( max - min ) + min );
    }
    
    clock_t start = clock ( );
    for ( int i = 0; i < iterations; ++i ) {
        switch ( type ) {
            case 0: {
                array_min_sequential ( a, length );
                break;
            }
            case 1: {
                array_min_vectorized ( a, length );
                break;
            }
        }
    }
    
    clock_t end = clock ( );
    
    double elapsed_in_ms = ( end - start ) * 1. / CLOCKS_PER_SEC * 1000;
    
    printf ( "ELAPSED (ms) = %lf\n", elapsed_in_ms );
    
    free ( a );
    
    return 0;
}