#include <stdlib.h>
#include <time.h>
#include <stdio.h>
#include <immintrin.h>

double array_sum_sequential ( double *a, int length ) {
    double sum = 0;
    
    for ( int i = 0; i < length; ++i ) {
        sum += a[i];
    }
    
    return sum;
}

double array_sum_vectorized ( double *a, int length ) {
    // __m256d can pack 4 doubles
    int overflow = length % 4;
    
    __m256d sum_vector = _mm256_set1_pd ( 0 );
    
    for ( int i = 0; i < ( length - overflow ); i += 4 ) {
        __m256d a_vector = _mm256_loadu_pd ( a + i );
        
        sum_vector = _mm256_add_pd ( a_vector, sum_vector );
    }
    
    double tmp[4] = { 0 };
    
    _mm256_storeu_pd ( tmp, sum_vector );
    
    double sum = tmp[0] + tmp[1] + tmp[2] + tmp[3];
    
    for ( int i = length - overflow; i < length; ++i ) {
        sum += a[i];
    }
    
    return sum;
}

int main ( int argc, char **argv ) {
    int type       = 1;
    int iterations = 10;
    int length     = 16;
    int min        = 0;
    int max        = 10;
    
    if ( argc >= 2 ) { type       = atoi ( argv[1] ); }
    if ( argc >= 3 ) { iterations = atoi ( argv[2] ); }
    if ( argc >= 4 ) { length     = atoi ( argv[3] ); }
    if ( argc >= 5 ) { min        = atoi ( argv[4] ); }
    if ( argc >= 6 ) { max        = atoi ( argv[5] ); }
    
    double *a = calloc ( length, sizeof ( double ) );
    
    if ( a == NULL ) {
        return 1;
    }
    
    for ( int i = 0; i < length; ++i ) {
        a[i] = 1. * rand ( ) / RAND_MAX * ( max - min ) + min;
    }
    
    clock_t start = clock ( );
    for ( int i = 0; i < iterations; ++i ) {
        switch ( type ) {
            case 0: {
                array_sum_sequential ( a, length );
                break;
            }
            case 1: {
                array_sum_vectorized ( a, length );
                break;
            }
        }
    }
    
    clock_t end = clock ( );
    
    double elapsed_in_ms = ( end - start ) * 1. / CLOCKS_PER_SEC * 1000;
    
    printf ( "ELAPSED (ms) = %lf\n", elapsed_in_ms );
    
    free ( a );
    
    return 0;
}